/*
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * CVS $Id: Value.java,v 1.10 2004/02/08 02:49:44 vgritsenko Exp $
 */

package org.apache.xindice.core.data;

import org.apache.xindice.util.XindiceRuntimeException;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

/**
 * Value is the primary base class for all data storing objects.
 * The content window of Value objects are immutable, but the
 * underlying byte array is not.
 *
 * @version CVS $Revision: 1.10 $, $Date: 2004/02/08 02:49:44 $
 */
public class Value implements Comparable {

    protected byte[] data = null;
    protected int pos = 0;
    protected int len = -1;


    public Value(Value value) {
        this.data = value.data;
        this.pos = value.pos;
        this.len = value.len;
    }

    public Value(byte[] data) {
        this.data = data;
        this.len = data.length;
    }

    public Value(byte[] data, int pos, int len) {
        this.data = data;
        this.pos = pos;
        this.len = len;
    }

    public Value(String data) {
        try {
            this.data = data.getBytes("utf-8");
            this.len = this.data.length;
        } catch (UnsupportedEncodingException e) {
            throw new XindiceRuntimeException("Java doesn't support UTF-8 encoding", e);
        }
    }

    /**
     * getData retrieves the data being stored by the Value as a byte array.
     *
     * @return The Data
     */
    public final byte[] getData() {
        if (len != data.length) {
            byte[] b = new byte[len];
            System.arraycopy(data, pos, b, 0, len);
            return b;
        } else {
            return data;
        }
    }

    /**
     * getLength retrieves the length of the data being stored by the Value.
     *
     * @return The Value length
     */
    public final int getLength() {
        return len;
    }

    /**
     * getInputStream returns an InputStream for the Value.
     *
     * @return An InputStream
     */
    public final InputStream getInputStream() {
        return new ByteArrayInputStream(data, pos, len);
    }

    /**
     * streamTo streams the content of the Value to an OutputStream.
     *
     * @param out the OutputStream
     */
    public final void streamTo(OutputStream out) throws IOException {
        out.write(data, pos, len);
    }

    public final void copyTo(byte[] tdata, int tpos) {
        System.arraycopy(data, pos, tdata, tpos, len);
    }

    public final void copyTo(byte[] tdata, int tpos, int len) {
        System.arraycopy(data, pos, tdata, tpos, len);
    }

    public final String toString() {
        try {
            return new String(data, pos, len, "utf-8");
        } catch (UnsupportedEncodingException e) {
            throw new XindiceRuntimeException("Java doesn't seem to support UTF-8!", e);
        }
    }

    public int hashCode() {
        return toString().hashCode();
    }

    public boolean equals(Value value) {
        return len == value.len ? compareTo(value) == 0 : false;
    }

    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj instanceof Value) {
            return equals((Value) obj);
        } else {
            return equals(new Value(obj.toString()));
        }
    }

    public final int compareTo(Value value) {
        byte[] ddata = value.data;
        int dpos = value.pos;
        int dlen = value.len;

        int stop = len > dlen ? dlen : len;

        for (int i = 0; i < stop; i++) {
            byte b1 = data[pos + i];
            byte b2 = ddata[dpos + i];

            if (b1 == b2)
                continue;
            else {
                short s1 = (short) (b1 >>> 0);
                short s2 = (short) (b2 >>> 0);
                return s1 > s2 ? (i + 1) : -(i + 1);
            }
        }

        if (len == dlen) {
            return 0;
        } else {
            return len > dlen ? stop + 1 : -(stop + 1);
        }
    }

    public final int compareTo(Object obj) {
        if (obj instanceof Value) {
            return compareTo((Value) obj);
        } else {
            return compareTo(new Value(obj.toString()));
        }
    }

    public final boolean startsWith(Value value) {
        if (len < value.len) {
            return false;
        }

        byte[] ddata = value.data;
        int dpos = value.pos;

        for (int i = 0; i < value.len; i++) {
            if (data[i + pos] != ddata[i + dpos]) {
                return false;
            }
        }

        return true;
    }
}
