/*
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * CVS $Id: StringUtilities.java,v 1.7 2004/02/08 02:59:39 vgritsenko Exp $
 */

package org.apache.xindice.util;

import java.util.StringTokenizer;

/**
 * StringUtilities provides a set of commonly used String parsing and
 * processing routines.
 *
 * @version CVS $Revision: 1.7 $, $Date: 2004/02/08 02:59:39 $
 */
public final class StringUtilities {

    private StringUtilities() {
    }

    /**
     * stringToInt quickly converts a String value to an int, returning
     * a default value if a conversion error occurs.
     *
     * @param value The value to convert
     * @param defvalue The default value to return
     * @return The converted value
     */
    public static int stringToInt(String value, int defvalue) {
        try {
            return Integer.parseInt(value);
        } catch (Exception e) {
            return defvalue;
        }
    }

    /**
     * stringToInt quickly converts a String value to an int, returning
     * 0 if a conversion error occurs.
     *
     * @param value The value to convert
     * @return The converted value
     */
    public static int stringToInt(String value) {
        return stringToInt(value, 0);
    }

    /**
     * leftJustify left-justifies a String value, space padding to width
     * characters if the string is less than width.
     *
     * @param value The value to left-justify
     * @param width The width to left-justify to
     * @return The left-justified value
     */
    public static String leftJustify(String value, int width) {
        StringBuffer sb = new StringBuffer(width);
        sb.append(value);
        for (int i = sb.length(); i < width; i++) {
            sb.append(' ');
        }
        
        if (sb.length() > width) {
            sb.setLength(width);
        }
        return sb.toString();
    }

    /**
     * rightJustify right-justifies a String value, space padding to
     * width characters if the string is less than width.
     *
     * @param value The value to right-justify
     * @param width The width to right-justify to
     * @return The right-justified value
     */
    public static String rightJustify(String value, int width) {
        StringBuffer sb = new StringBuffer(width);
        sb.append(value);
        for (int i = sb.length(); i < width; i++) {
            sb.insert(0, ' ');
        }
        
        if (sb.length() > width) {
            return sb.toString().substring(sb.length() - width);
        } else {
            return sb.toString();
        }
    }

    /**
     * parseBuffer parses through a String buffer and produces a
     * Map table using the provided delimiters.  This method can also
     * urldecode values and produce newline-delimited multi-value properties if
     * multiple values for the same key are detected in the buffer.
     *
     * @param retprops The Map to populate
     * @param inbuffer The String buffer to parse
     * @param delimiters The delimiters to use
     * @param urldecode Whether or not to URL-decode the values
     * @param multivalue Whether or not to parse multiple values
     */
    /*
   public static void parseBuffer(Map retprops, String inbuffer, String delimiters, boolean urldecode, boolean multivalue) {
      if ( inbuffer.length() == 0 )
         return;
      int idx = 0;
      String line;
      String key;
      String value;
      String prop;
      String temp;
      StringTokenizer st = new StringTokenizer(inbuffer, delimiters, false);
      while ( st.hasMoreTokens() ) {
         line = st.nextToken();
         idx = line.indexOf('=');
         if ( idx != -1 ) {
            key = line.substring(0, idx).trim();
            try {
               value = urldecode ? URLDecoder.decode(line.substring(idx+1).trim())
                                 : line.substring(idx+1).trim();
            }
            catch ( Exception e ) {
               value = line.substring(idx+1).trim();
            }
            if ( multivalue ) {
               temp = (String)retprops.get(key);
               if ( temp != null && temp.length() > 0 )
                  value = temp + "\u0001" + value;
               retprops.put(key, value);
            }
            else
               retprops.put(key, value);
         }
      }
   }
    */

    /**
     * findWhiteSpace scans a String value for whitespace characters
     * from a specified starting position and returns the location of the first
     * whitespace character found.
     *
     * @param value The value to scan
     * @param start The starting position
     * @return The first whitespace location
     */
    public static int findWhiteSpace(String value, int start) {
        char[] chars = value.toCharArray();
        for (int i = 0; i < chars.length; i++) {
            if (chars[i] <= 32) {
                return i;
            }
        }
        return -1;
    }

    /**
     * findWhiteSpace scans a String value for whitespace characters
     * and returns the location of the first whitespace character found.
     *
     * @param value The value to scan
     * @return The first whitespace location
     */
    public static int findWhiteSpace(String value) {
        return findWhiteSpace(value, 0);
    }

    /**
     * javaEncode converts a String value to a Java-printable String.
     * All non-printable characters, such as newline and backspace are converted
     * to their respective escape sequences.  This method does not convert
     * UNICODE characters.
     *
     * @param value The value to convert
     * @return The converted value
     */
    public static String javaEncode(String value) {
        StringTokenizer st = new StringTokenizer(value, "\b\t\n\f\r\"\'\\", true);
        StringBuffer sb = new StringBuffer(value.length());
        String token;
        while (st.hasMoreTokens()) {
            token = st.nextToken();
            if (token.equals("\b")) {
                sb.append("\\b");
            } else if (token.equals("\t")) {
                sb.append("\\t");
            } if (token.equals("\n")) {
                sb.append("\\n");
            } else if (token.equals("\f")) {
                sb.append("\\f");
            } else if (token.equals("\r")) {
                sb.append("\\r");
            } else if (token.equals("\"")) {
                sb.append("\\\"");
            } else if (token.equals("\'")) {
                sb.append("\\\'");
            } else if (token.equals("\\")) {
                sb.append("\\\\");
            } else {
                sb.append(token);
            }
        }
        return sb.toString();
    }
}
