/*
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * CVS $Id: XMLCompressedOutput.java,v 1.6 2004/02/08 03:50:13 vgritsenko Exp $
 */

package org.apache.xindice.xml;

import java.io.DataOutputStream;
import java.io.IOException;
import java.io.OutputStream;

/**
 * XMLCompressedInput is an OutputStream extension that provides functions for
 * writing types to a Xindice Compressed XML Stream.
 *
 * @version CVS $Revision: 1.6 $, $Date: 2004/02/08 03:50:13 $
 */
public class XMLCompressedOutput extends DataOutputStream {
    protected SymbolTable st = null;

    public XMLCompressedOutput(OutputStream os, SymbolTable st) {
        super(os);
        this.st = st;
    }

    /**
     * getSizeType returns the optimal size type for the specifed
     * length value.
     *
     * @see org.apache.xindice.xml.Signatures
     *
     * @param length The length to evaluate
     * @return The optimal size
     */
    protected final byte getSizeType(long length) {
        if (length > Short.MAX_VALUE) {
            return Signatures.IntContent;
        } else if (length > Byte.MAX_VALUE) {
            return Signatures.ShortContent;
        } else if (length > 0) {
            return Signatures.ByteContent;
        } else {
            return Signatures.NoContent;
        }
    }

    /**
     * getSizeSize returns the size (in bytes) of the specified
     * size type.
     *
     * @see org.apache.xindice.xml.Signatures
     *
     * @param sizeType The size type
     * @return The size in bytes
     */
    protected final int getSizeSize(int sizeType) {
        switch (sizeType) {
            case Signatures.IntContent:
                return 4;
            case Signatures.ShortContent:
                return 2;
            case Signatures.ByteContent:
                return 1;
            default:
                return 0;
        }
    }

    /**
     * writeSize writes the size to the output stream based on the
     * specified sizeType.
     *
     * @see org.apache.xindice.xml.Signatures
     *
     * @param sizeType The size type
     * @param size The size
     * @return The size in bytes written
     * @throws IOException If the write failed
     */
    protected final int writeSize(int sizeType, int size) throws IOException {
        switch (sizeType) {

            case Signatures.IntContent:
                writeInt(size);
                return 4;

            case Signatures.ShortContent:
                writeShort((short) size);
                return 2;

            case Signatures.ByteContent:
                writeByte((byte) size);
                return 1;

            default:
                return 0;
        }
    }
}


